/*
 * Power up hardware initialization for IXP42x processors.
 */

#include <linux/linkage.h>
#include <asm/mach-types.h>

.section        ".start", "ax"

_IXP42x_hw_init:

	/* Execute from flash address space */
	add	pc, pc, #0x50000000		@ Flash is at fixed address
	nop

	/* Init the primary UART */
	mov	r0, #0xc8000000			@ UART address base
	mov	r1, #0x83			@ Enable mode 8n1
	str	r1, [r0,#12]
	mov	r1, #0x08			@ Set 115200 baud
	str	r1, [r0,#0]
	mov	r1, #0x00
	str	r1, [r0,#4]
	mov	r1, #0x03
	str	r1, [r0,#12]
	mov	r1, #0x40
	str	r1, [r0,#4]

	/* Speed up the FLASH chip select timing */
	mov	r0, #0xc4000000			@ Timing and CTRL reg
	mov	r1, #0x80000000			@ Set fast access timing
	orr	r1, r1, #0x00c00000
	orr	r1, r1, #0x00003c00
	orr	r1, r1, #0x0000004b
	str	r1, [r0]			@ Enable the fast timing

	/* Turn on big-endian mode */
	mrc	p15, 0, r0, c1, c0, 0
	orr	r0, r0, #0x80
	mcr	p15, 0, r0, c1, c0, 0

	mrc	p15, 0, r0, c2, c0, 0		@ cpwait - wait for CP15
	mov	r0, r0				@ As per IXP4xx doc
	sub	pc, pc, #4

	/* Init and startup SDRAM controller */
	mov	r0, #0xcc000000			@ SDRAM controller registers
	mov	r1, #0x18			@ 32MB config value
	str	r1, [r0,#0]			@ Set SDR_CONFIG reg
	mov	r1, #0x03			@ SDRAM NOP command
	str	r1, [r0,#8]			@ Issue command

	/* 200us delay required */
	mov	r1, #0x8000			@ Delay counter
1:	subs	r1, r1, #1			@ Decrement counter
	bne	1b

	mov	r1, #0x02			@ Pre-charge all command
	str	r1, [r0,#8]			@ Issue command
	mov	r3, #8				@ Do 8 auto-refresh cycles
1:	mov	r1, #0x04			@ Auto-refresh command
	str	r1, [r0,#8]			@ Issue command
	ldr	r2, [r0,#8]			@ Reg read back (delay)
	subs	r3, r3, #1			@ Decrement loop counter
	bne	1b

	mov	r1, #0x01			@ Mode register set command
	str	r1, [r0,#8]			@ Issue command
	mov	r1, #0x384			@ Refresh count value
	str	r1, [r0,#4]			@ Start refresh cycles
	mov	r1, #0x06			@ Normal operation command
	str	r1, [r0,#8]			@ Issue command

	/* Switch RAM to be based at address 0 */
	mov	r0, #0xc4000000			@ Expansion Bus config address
	ldr	r1, [r0,#32]			@ Get current EXP_CBFG0 value
	and	r1, r1, #0x7fffffff		@ Map RAM from low address
	str	r1, [r0,#32]			@ Enable mapping

	mov	sp, #(32*1024*1024)		@ Set tmp stack to end of mem
	sub	sp, sp, #4

	/* Enable the CPU level 1 cache */
	mrc	p15, 0, r0, c1, c0, 0		@ Enable cache
	orr	r0, r0, #0x00001000
	orr	r0, r0, #0x0000000c
	mcr	p15, 0, r0, c1, c0, 0

	mrc	p15, 0, r0, c2, c0, 0		@ cpwait - wait for CP15
	mov	r0, r0				@ As per IXP4xx doc
	sub	pc, pc, #4

	/* Copy this zImage from flash into RAM */
	mov	r0, #0x50000000			@ Start of flash
	mov	r1, #0x00100000			@ Copy to 1MB in RAM
	add	r2, r1, #0x00120000		@ Copy 1.1MB worth of zImage
1:	ldmia	r0!, {r3-r6}			@ Read some zImage
	stmia	r1!, {r3-r6}			@ Put some zImage
	cmp	r1, r2				@ Check if all done
	bne	1b

	/* Execute from RAM now */
	mov	r0, #0x50000000			@ Flash address
	sub	r0, r0, #0x00100000		@ Remove RAM load offset
	sub	pc, pc, r0			@ Jump into RAM
	nop

	/* Setup up PCI bus clock */
	mov	r0, #0xc8000000			@ Get address of GPIO regs
	orr	r0, r0, #0x00004000
	ldr	r1, [r0,#0x18]			@ Get Clock CTRL reg
	mov	r2, #0xff000000
	orr	r2, r2, #0x00ff0000
	and	r1, r1, r2			@ Mask out lower 16 bits
	orr	r1, #0x000000ff			@ Enable PCI clock
	orr	r1, #0x00000100
	str	r1, [r0,#0x18]			@ Start the PCI clock output
	/* Initialize the GPIO lines */
	mov	r1, #0x00000003			@ Set IOs as required
	orr	r1, r1, #0x00001b00		@ Pull PCI bus out of reset too
	str	r1, [r0,#0x04]
	mov	r1, #0x000000fc			@ Set GPIO CTRL as required
	orr	r1, r1, #0x00002400
	str	r1, [r0, #0x00]

	/* Set the ARM machine type ID */
	mov	r7, #MACH_TYPE_SG565 & 0xff	@ Hard code machine ID
	orr	r7, r7, #MACH_TYPE_SG565 & 0xff00
	mov	r8, #0x100			@ Dummy ATAGS

